// ZoomBmpDoc.cpp : implementation of the CZoomBmpDoc class
//

#include "stdafx.h"
#include "ZoomBmp.h"

#include "ZoomBmpDoc.h"
#include "MainFrm.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CZoomBmpDoc

IMPLEMENT_DYNCREATE(CZoomBmpDoc, CDocument)

BEGIN_MESSAGE_MAP(CZoomBmpDoc, CDocument)
	//{{AFX_MSG_MAP(CZoomBmpDoc)
	ON_CBN_KILLFOCUS(IDC_ZOOM_COMBO, OnSelEndOk)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CZoomBmpDoc construction/destruction

CZoomBmpDoc::CZoomBmpDoc() : m_hDIB(NULL), m_hSrcDIB(NULL)
{
}

CZoomBmpDoc::~CZoomBmpDoc()
{
}

BOOL CZoomBmpDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

	return TRUE;
}



/////////////////////////////////////////////////////////////////////////////
// CZoomBmpDoc serialization

void CZoomBmpDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		// TODO: add storing code here
	}
	else
	{
		// TODO: add loading code here
	}
}

/////////////////////////////////////////////////////////////////////////////
// CZoomBmpDoc diagnostics

#ifdef _DEBUG
void CZoomBmpDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CZoomBmpDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CZoomBmpDoc commands

BOOL CZoomBmpDoc::OnOpenDocument(LPCTSTR lpszPathName) 
{
	// clean SDI doc
	DeleteContents();	

	CFile file(lpszPathName, CFile::modeRead | CFile::shareDenyWrite);
	m_hSrcDIB = ::ReadDIBFile(file); // read bitmap

	if (m_hSrcDIB == NULL)  // not DIB file
		return FALSE;

	LPSTR lpDIB = (LPSTR) ::GlobalLock((HGLOBAL) m_hSrcDIB);
	int numColors = ::DIBNumColors(lpDIB);
	::GlobalUnlock((HGLOBAL) m_hSrcDIB);

	if (numColors != 2) { 
		::GlobalFree((HGLOBAL) m_hSrcDIB);
		m_hSrcDIB = NULL;
		AfxMessageBox("Only monochrome bitmaps are supported");
		return FALSE;
	}

	return TRUE;
}

void CZoomBmpDoc::DeleteContents() 
{
	if (m_hDIB != NULL)
		::GlobalFree((HGLOBAL) m_hDIB);
	
	if (m_hSrcDIB != NULL)
		::GlobalFree((HGLOBAL) m_hSrcDIB);

	CDocument::DeleteContents();
}

void CZoomBmpDoc::OnSelEndOk()
{
	CControlBar* pBar = ((CFrameWnd*)AfxGetMainWnd())->GetControlBar(AFX_IDW_TOOLBAR);
	CComboBox& cb = ((CMainToolBar*)pBar)->m_cbZoom;
	CString txt;
	cb.GetWindowText(txt);
	double zoom = 100.0 / atof(txt); // txt is a number in %

	ZoomBmp(zoom);

	UpdateAllViews(NULL);
}


///////////////////////////////////////////////////////////////////
// Function name	: CZoomBmpDoc::ZoomBmp
// Description	    : creates a new bitmap which is a grayscale 
//					  zoomed out version of the original
// Return type		: void 
// Argument         : double zoom
///////////////////////////////////////////////////////////////////
void CZoomBmpDoc::ZoomBmp(double zoom)
{
	if (m_hSrcDIB == NULL) // nothing to do
		return;

	if (zoom < 1) // foolproof
		return;

	// free previous bitmap to save some memory
	if (m_hDIB) 
		::GlobalFree((HGLOBAL) m_hDIB);

	LPSTR pSrcDIB = (LPSTR) ::GlobalLock((HGLOBAL) m_hSrcDIB);
	BITMAPINFOHEADER& bmihSrc = *(BITMAPINFOHEADER*)pSrcDIB;
	ASSERT(bmihSrc.biBitCount == 1);
	LPSTR pSrcBits = (LPSTR) ::FindDIBBits(pSrcDIB);

	BITMAPINFOHEADER bmihDst = bmihSrc;
	// change params 
	bmihDst.biWidth = (LONG)(bmihDst.biWidth / zoom + 0.5);
	bmihDst.biHeight = (LONG)(bmihDst.biHeight / zoom + 0.5);
	bmihDst.biBitCount = 8; // grayscale in any case
	bmihDst.biClrUsed = 0; // to be on the safe side

	// prepare destination bitmap
	DWORD dwDIBSize = sizeof(bmihDst) + ::PaletteSize((LPSTR)&bmihDst) +
		WIDTHBYTES(bmihDst.biWidth * bmihDst.biBitCount) * bmihDst.biHeight;
	bmihDst.biSizeImage = dwDIBSize;

	// allocate space for the new bitmap
	HDIB hDstDIB = (HDIB) ::GlobalAlloc(GMEM_MOVEABLE | GMEM_ZEROINIT, dwDIBSize);
	if (hDstDIB == 0) {
		::GlobalUnlock((HGLOBAL) m_hSrcDIB);
		return;
	}
	LPSTR pDstDIB = (LPSTR) ::GlobalLock((HGLOBAL) hDstDIB);

	// copy header
	memcpy(pDstDIB, &bmihDst, sizeof(bmihDst));
	// prepare palette
	for (int i=0; i < (1 << bmihDst.biBitCount); i++) {
		RGBQUAD& palEntry = *(RGBQUAD*)(pDstDIB + sizeof(bmihDst) + i * sizeof(RGBQUAD));
		palEntry.rgbRed = palEntry.rgbGreen = palEntry.rgbBlue = i;
	}
	LPSTR pDstBits = (LPSTR) ::FindDIBBits(pDstDIB);

	// now fill the bits

	LPSTR curSrcLineBits, curDstLineBits;
	int j, k;
	int scale = (int)(zoom + 0.5); // integer scale factor, i.e. 1:5
	int hBase, vBase;
	unsigned char value;

	// for every _scale_ lines in a source bitmap we will get one line
	// in the destination bitmap. Similarly for _scale_ columns in the 
	// source we'll obtain one destination column.
	for (int strip=0; strip < bmihDst.biHeight; strip++) { // for every dst line

		curDstLineBits = pDstBits + strip * WIDTHBYTES(bmihDst.biWidth * bmihDst.biBitCount); 
		vBase = int(strip * zoom + 0.5);

		for (i=0; i < scale; i++) {  // accumulate _scale_ rows

			curSrcLineBits = pSrcBits + (vBase + i) * WIDTHBYTES(bmihSrc.biWidth * bmihSrc.biBitCount); 

			// prepare horizontally condensed lines for this strip
			for (j=0; j < bmihDst.biWidth; j++) {	// for all bits in line

				hBase = int(j * zoom + 0.5); // mapped index on source

				for (k=0; k < scale; k++) { // accumulate _scale_ columns
					value = (curSrcLineBits[(hBase+k)/8] & (1 << (7 - (hBase+k)%8))) ? 0xff : 0;
					curDstLineBits[j] += value / scale / scale; // main accumulator
				}
			}
		}
	}

	// unlock memory
	::GlobalUnlock((HGLOBAL) m_hSrcDIB);
	::GlobalUnlock((HGLOBAL) hDstDIB);

	// replace bitmap
	m_hDIB = hDstDIB;
}
